@@ -5,6 +5,8 @@ from __future__ import annotations
 from typing import TYPE_CHECKING
 from typing import Optional
 
+from .token import TOKEN_EOF
+
 if TYPE_CHECKING:
     from .token import Token
 
@@ -22,13 +24,69 @@ class JSONPathError(Exception):
         self.token: Optional[Token] = token
 
     def __str__(self) -> str:
-        msg = super().__str__()
+        return self.detailed_message()
 
+    def detailed_message(self) -> str:
+        """Return an error message formatted with extra context info."""
         if not self.token:
-            return msg
+            return super().__str__()
 
-        line, column = self.token.position()
-        return f"{msg}, line {line}, column {column}"
+        lineno, col, _prev, current, _next = self._error_context(
+            self.token.path, self.token.index
+        )
+
+        if self.token.kind == TOKEN_EOF:
+            col = len(current)
+
+        pad = " " * len(str(lineno))
+        length = len(self.token.value)
+        pointer = (" " * col) + ("^" * max(length, 1))
+
+        return (
+            f"{self.message}\n"
+            f"{pad} -> {self.token.path!r} {lineno}:{col}\n"
+            f"{pad} |\n"
+            f"{lineno} | {current}\n"
+            f"{pad} | {pointer} {self.message}\n"
+        )
+
+    @property
+    def message(self) -> object:
+        """The exception's error message if one was given."""
+        if self.args:
+            return self.args[0]
+        return None
+
+    def _error_context(self, text: str, index: int) -> tuple[int, int, str, str, str]:
+        lines = text.splitlines(keepends=True)
+        cumulative_length = 0
+        target_line_index = -1
+
+        for i, line in enumerate(lines):
+            cumulative_length += len(line)
+            if index < cumulative_length:
+                target_line_index = i
+                break
+
+        if target_line_index == -1:
+            raise ValueError("index is out of bounds for the given string")
+
+        # Line number (1-based)
+        line_number = target_line_index + 1
+        # Column number within the line
+        column_number = index - (cumulative_length - len(lines[target_line_index]))
+
+        previous_line = (
+            lines[target_line_index - 1].rstrip() if target_line_index > 0 else ""
+        )
+        current_line = lines[target_line_index].rstrip()
+        next_line = (
+            lines[target_line_index + 1].rstrip()
+            if target_line_index < len(lines) - 1
+            else ""
+        )
+
+        return line_number, column_number, previous_line, current_line, next_line
 
 
 class JSONPathSyntaxError(JSONPathError):
@@ -77,6 +135,19 @@ class JSONPathNameError(JSONPathError):
         self.token = token
 
 
+class JSONPathRecursionError(JSONPathError):
+    """An exception raised when the maximum recursion depth is reached.
+
+    Arguments:
+        args: Arguments passed to `Exception`.
+        token: The token that caused the error.
+    """
+
+    def __init__(self, *args: object, token: Token) -> None:
+        super().__init__(*args)
+        self.token = token
+
+
 class JSONPointerError(Exception):
     """Base class for all JSON Pointer errors."""
 
