@@ -1,4 +1,5 @@
 """JSON Patch, as per RFC 6902."""
+
 from __future__ import annotations
 
 import copy
@@ -15,14 +16,15 @@ from typing import MutableSequence
 from typing import TypeVar
 from typing import Union
 
-from .exceptions import JSONPatchError
-from .exceptions import JSONPatchTestFailure
-from .exceptions import JSONPointerError
-from .exceptions import JSONPointerIndexError
-from .exceptions import JSONPointerKeyError
-from .exceptions import JSONPointerTypeError
-from .pointer import UNDEFINED
-from .pointer import JSONPointer
+from jsonpath._data import load_data
+from jsonpath.exceptions import JSONPatchError
+from jsonpath.exceptions import JSONPatchTestFailure
+from jsonpath.exceptions import JSONPointerError
+from jsonpath.exceptions import JSONPointerIndexError
+from jsonpath.exceptions import JSONPointerKeyError
+from jsonpath.exceptions import JSONPointerTypeError
+from jsonpath.pointer import UNDEFINED
+from jsonpath.pointer import JSONPointer
 
 
 class Op(ABC):
@@ -72,7 +74,7 @@ class OpAdd(Op):
             else:
                 parent.insert(int(target), self.value)
         elif isinstance(parent, MutableMapping):
-            parent[target] = self.value
+            parent[str(target)] = self.value
         else:
             raise JSONPatchError(
                 f"unexpected operation on {parent.__class__.__name__!r}"
@@ -84,6 +86,78 @@ class OpAdd(Op):
         return {"op": self.name, "path": str(self.path), "value": self.value}
 
 
+class OpAddNe(OpAdd):
+    """A non-standard _add if not exists_ operation.
+
+    This is like _OpAdd_, but only adds object/dict keys/values if they key does
+    not already exist.
+
+    **New in version 1.2.0**
+    """
+
+    __slots__ = ("path", "value")
+
+    name = "addne"
+
+    def apply(
+        self, data: Union[MutableSequence[object], MutableMapping[str, object]]
+    ) -> Union[MutableSequence[object], MutableMapping[str, object]]:
+        """Apply this patch operation to _data_."""
+        parent, obj = self.path.resolve_parent(data)
+        if parent is None:
+            # Replace the root object.
+            # The following op, if any, will raise a JSONPatchError if needed.
+            return self.value  # type: ignore
+
+        target = self.path.parts[-1]
+        if isinstance(parent, MutableSequence):
+            if obj is UNDEFINED:
+                parent.append(self.value)
+            else:
+                parent.insert(int(target), self.value)
+        elif isinstance(parent, MutableMapping) and target not in parent:
+            parent[target] = self.value
+        return data
+
+
+class OpAddAp(OpAdd):
+    """A non-standard add operation that appends to arrays/lists .
+
+    This is like _OpAdd_, but assumes an index of "-" if the path can not
+    be resolved rather than raising a JSONPatchError.
+
+    **New in version 1.2.0**
+    """
+
+    __slots__ = ("path", "value")
+
+    name = "addap"
+
+    def apply(
+        self, data: Union[MutableSequence[object], MutableMapping[str, object]]
+    ) -> Union[MutableSequence[object], MutableMapping[str, object]]:
+        """Apply this patch operation to _data_."""
+        parent, obj = self.path.resolve_parent(data)
+        if parent is None:
+            # Replace the root object.
+            # The following op, if any, will raise a JSONPatchError if needed.
+            return self.value  # type: ignore
+
+        target = self.path.parts[-1]
+        if isinstance(parent, MutableSequence):
+            if obj is UNDEFINED:
+                parent.append(self.value)
+            else:
+                parent.insert(int(target), self.value)
+        elif isinstance(parent, MutableMapping):
+            parent[target] = self.value
+        else:
+            raise JSONPatchError(
+                f"unexpected operation on {parent.__class__.__name__!r}"
+            )
+        return data
+
+
 class OpRemove(Op):
     """The JSON Patch _remove_ operation."""
 
@@ -109,7 +183,7 @@ class OpRemove(Op):
         elif isinstance(parent, MutableMapping):
             if obj is UNDEFINED:
                 raise JSONPatchError("can't remove nonexistent property")
-            del parent[self.path.parts[-1]]
+            del parent[str(self.path.parts[-1])]
         else:
             raise JSONPatchError(
                 f"unexpected operation on {parent.__class__.__name__!r}"
@@ -147,7 +221,7 @@ class OpReplace(Op):
         elif isinstance(parent, MutableMapping):
             if obj is UNDEFINED:
                 raise JSONPatchError("can't replace nonexistent property")
-            parent[self.path.parts[-1]] = self.value
+            parent[str(self.path.parts[-1])] = self.value
         else:
             raise JSONPatchError(
                 f"unexpected operation on {parent.__class__.__name__!r}"
@@ -185,7 +259,7 @@ class OpMove(Op):
         if isinstance(source_parent, MutableSequence):
             del source_parent[int(self.source.parts[-1])]
         if isinstance(source_parent, MutableMapping):
-            del source_parent[self.source.parts[-1]]
+            del source_parent[str(self.source.parts[-1])]
 
         dest_parent, _ = self.dest.resolve_parent(data)
 
@@ -196,7 +270,7 @@ class OpMove(Op):
         if isinstance(dest_parent, MutableSequence):
             dest_parent.insert(int(self.dest.parts[-1]), source_obj)
         elif isinstance(dest_parent, MutableMapping):
-            dest_parent[self.dest.parts[-1]] = source_obj
+            dest_parent[str(self.dest.parts[-1])] = source_obj
         else:
             raise JSONPatchError(
                 f"unexpected operation on {dest_parent.__class__.__name__!r}"
@@ -238,7 +312,7 @@ class OpCopy(Op):
         if isinstance(dest_parent, MutableSequence):
             dest_parent.insert(int(self.dest.parts[-1]), copy.deepcopy(source_obj))
         elif isinstance(dest_parent, MutableMapping):
-            dest_parent[self.dest.parts[-1]] = copy.deepcopy(source_obj)
+            dest_parent[str(self.dest.parts[-1])] = copy.deepcopy(source_obj)
         else:
             raise JSONPatchError(
                 f"unexpected operation on {dest_parent.__class__.__name__!r}"
@@ -339,6 +413,16 @@ class JSONPatch:
                     path=self._op_pointer(operation, "path", "add", i),
                     value=self._op_value(operation, "value", "add", i),
                 )
+            elif op == "addne":
+                self.addne(
+                    path=self._op_pointer(operation, "path", "addne", i),
+                    value=self._op_value(operation, "value", "addne", i),
+                )
+            elif op == "addap":
+                self.addap(
+                    path=self._op_pointer(operation, "path", "addap", i),
+                    value=self._op_value(operation, "value", "addap", i),
+                )
             elif op == "remove":
                 self.remove(path=self._op_pointer(operation, "path", "add", i))
             elif op == "replace":
@@ -382,9 +466,12 @@ class JSONPatch:
                 f"({op}:{i})"
             )
 
-        return JSONPointer(
-            pointer, unicode_escape=self.unicode_escape, uri_decode=self.uri_decode
-        )
+        try:
+            return JSONPointer(
+                pointer, unicode_escape=self.unicode_escape, uri_decode=self.uri_decode
+            )
+        except JSONPointerError as err:
+            raise JSONPatchError(f"{err} ({op}:{i})") from err
 
     def _op_value(
         self, operation: Mapping[str, object], key: str, op: str, i: int
@@ -420,6 +507,38 @@ class JSONPatch:
         self.ops.append(OpAdd(path=pointer, value=value))
         return self
 
+    def addne(self: Self, path: Union[str, JSONPointer], value: object) -> Self:
+        """Append an _addne_ operation to this patch.
+
+        Arguments:
+            path: A string representation of a JSON Pointer, or one that has
+                already been parsed.
+            value: The object to add.
+
+        Returns:
+            This `JSONPatch` instance, so we can build a JSON Patch by chaining
+                calls to JSON Patch operation methods.
+        """
+        pointer = self._ensure_pointer(path)
+        self.ops.append(OpAddNe(path=pointer, value=value))
+        return self
+
+    def addap(self: Self, path: Union[str, JSONPointer], value: object) -> Self:
+        """Append an _addap_ operation to this patch.
+
+        Arguments:
+            path: A string representation of a JSON Pointer, or one that has
+                already been parsed.
+            value: The object to add.
+
+        Returns:
+            This `JSONPatch` instance, so we can build a JSON Patch by chaining
+                calls to JSON Patch operation methods.
+        """
+        pointer = self._ensure_pointer(path)
+        self.ops.append(OpAddAp(path=pointer, value=value))
+        return self
+
     def remove(self: Self, path: Union[str, JSONPointer]) -> Self:
         """Append a _remove_ operation to this patch.
 
@@ -497,7 +616,7 @@ class JSONPatch:
         Arguments:
             path: A string representation of a JSON Pointer, or one that has
                 already been parsed.
-            value: The object to add.
+            value: The object to test.
 
         Returns:
             This `JSONPatch` instance, so we can build a JSON Patch by chaining
@@ -510,14 +629,14 @@ class JSONPatch:
     def apply(
         self,
         data: Union[str, IOBase, MutableSequence[object], MutableMapping[str, object]],
-    ) -> Union[MutableSequence[object], MutableMapping[str, object]]:
+    ) -> object:
         """Apply all operations from this patch to _data_.
 
         If _data_ is a string or file-like object, it will be loaded with
         _json.loads_. Otherwise _data_ should be a JSON-like data structure and
-        will be modified in-place.
+        will be modified in place.
 
-        When modifying _data_ in-place, we return modified data too. This is
+        When modifying _data_ in place, we return modified data too. This is
         to allow for replacing _data's_ root element, which is allowed by some
         patch operations.
 
@@ -532,14 +651,7 @@ class JSONPatch:
             JSONPatchTestFailure: When a _test_ operation does not pass.
                 `JSONPatchTestFailure` is a subclass of `JSONPatchError`.
         """
-        if isinstance(data, str):
-            _data: Union[
-                MutableSequence[object], MutableMapping[str, object]
-            ] = json.loads(data)
-        elif isinstance(data, IOBase):
-            _data = json.loads(data.read())
-        else:
-            _data = data
+        _data = load_data(data)
 
         for i, op in enumerate(self.ops):
             try:
@@ -547,19 +659,14 @@ class JSONPatch:
             except JSONPatchTestFailure as err:
                 raise JSONPatchTestFailure(f"test failed ({op.name}:{i})") from err
             except JSONPointerKeyError as err:
-                raise JSONPatchError(
-                    f"pointer key error {err} ({op.name}:{i})"
-                ) from err
+                raise JSONPatchError(f"{err} ({op.name}:{i})") from err
             except JSONPointerIndexError as err:
-                raise JSONPatchError(
-                    f"pointer index error {err} ({op.name}:{i})"
-                ) from err
+                raise JSONPatchError(f"{err} ({op.name}:{i})") from err
             except JSONPointerTypeError as err:
-                raise JSONPatchError(
-                    f"pointer type error {err} ({op.name}:{i})"
-                ) from err
+                raise JSONPatchError(f"{err} ({op.name}:{i})") from err
             except (JSONPointerError, JSONPatchError) as err:
                 raise JSONPatchError(f"{err} ({op.name}:{i})") from err
+
         return _data
 
     def asdicts(self) -> List[Dict[str, object]]:
